* 5-6 Feb 02: a general program for algebraic enumeration of the 
*             Langford sequences.
* 7-8 Feb 02: some optimizations and real*8 fudge (have used it
*             successfully up to n=16)
* 9 Feb 02:   arbitrary length integer arithmetic replaces real*8
* 11 Feb 02:  factors are grouped automatically in MKTERM
* 12 Feb 02:  array(*) for arrays passed to MKTERM & dependent routines 
      program lang13
      implicit none
      integer maxn, max2n, maxlen
      parameter (maxn = 20, max2n = (2*maxn), maxlen = 50)
      integer n, twon, x(max2n), sum(maxlen), term(maxlen), ios
      real t1, t2

      n = -1
      ios = -1
      do while(n.lt.3 .or. n.gt.maxn .or. ios.ne.0)
         print'(a,i2,a)','Order of Langford problem (up to ',
     &        maxn,'):' 
         read(*,fmt=*,iostat=ios) n
      end do

      call clock@(t1)
      twon = 2*n
      call lang(x,twon,sum,term,maxlen)
      call prnsum(n,sum,maxlen)
      call clock@(t2)
      print'("Time (seconds):",f9.2)', t2-t1

      end


      subroutine lang(x,twon,sum,term,maxlen)
      implicit none
      integer twon, maxlen
      integer x(twon), sum(maxlen), term(maxlen)
      integer n, dep, sgn, lentrm, i, m, depred
* definitions for two statement-functions:
      integer nexmov
      logical zero, nomore
      nomore(m) = x(m).eq.0
      nexmov(m) = mod(m+2,3)-1

      n = twon/2

* depth at which reduction of SUM occurs:
      depred = max(twon-13,1)

* initialize variables
      dep = 2
      sgn = 1
      x(1) = 1
      x(2) = 0
      do i = 1, maxlen
         sum(i) = 0
      end do

* this "tree search" generates all combinations of X(I) = +1 or -1.
* efficiency isn't crucial, as most of the time is spent in MKTERM

 10   x(dep) = nexmov(x(dep))
      if (nomore(dep)) then 
* backtrack to previous depth      
         dep = dep-1
         sgn = -sgn
* would normally return from here:         
         if (dep .eq. 1) then
            call final(sum,maxlen,twon)
            return
         end if
* reduce the terms in SUM to the range -BASE+1 to BASE-1         
         if (dep .eq. depred) call reduce(sum,maxlen)
      else 
* SGN accumulates the product of the X(I)...
         sgn = sgn*x(dep)
         if (dep .eq. twon) then
* evaluate new TERM ... include in SUM
            call mkterm(x,twon,term,lentrm,zero)
            if (.not.zero) call add(sum,term,sgn,lentrm)
         else 
* go deeper into tree
            dep = dep+1
            x(dep) = 0
         end if
      end if
      go to 10

      end
      

      subroutine mkterm(x,twon,term,big,zero)
      implicit none
      integer twon, big
      integer x(*), term(*)
      logical zero
      
      integer i, j, n, maxn, base, prod
      parameter (maxn = 20, base = 10000)
      integer fac(maxn)
      
      n = twon/2
* the slow part: calculating the factors in the generating function
      do i = n, 1, -1
         fac(i) = 0
         do j = twon-i-1, 1, -1
            fac(i) = fac(i) + x(j)*x(j+i+1)
         end do
         if (fac(i) .eq. 0) then
            zero = .true.
            return
         end if
      end do

      zero = .false.
      big = 1
* we could just multiply the factors into TERM
C      term(1) = 1
C      do i = n, 1, -1
C         call mult(term,big,fac(i))
C      end do
C      return
* but instead accumulate products of factors, to a limit of BASE-1
* (so as to minimize number of calls to MULT!)
      prod = fac(n)
* assumption here is that 2n is not too large -- start from PROD=1
* and I=N in general!  But this is all right up to N=20.
      do i = n-1, 1, -1
         prod = prod*fac(i)
         if (abs(prod).ge.base) then
            term(1) = prod/fac(i)
            prod = fac(i)
            j = i
            go to 10
         end if
      end do

* that loop DOES sometimes complete, so...
      term(1) = 1
      call mult(term,big,prod)
      return

 10   do i = j-1, 1, -1
* same assumption as above:
         prod = prod*fac(i)
         if (abs(prod).ge.base) then
            prod = prod/fac(i)
            call mult(term,big,prod)
            prod = fac(i)
         end if
      end do

      call mult(term,big,prod)

      end
      
*** routines for dealing with big integers ***
      
      subroutine mult(num,len,fac)
      implicit none
      integer len, fac
      integer num(*)

      integer base
      parameter (base = 10000)
      integer carry, i, temp
      
* multiplication to a given BASE: it doesn't matter if some
* of the NUM(I) are negative
      carry = 0
      do i = 1, len
         temp = num(i)*fac + carry
         num(i) = mod(temp,base)
         carry = temp/base
      end do
      
      if (carry .ne. 0) then
         len = len+1
* feeling confident enough to omit this check:
C         if (len .gt. maxlen) then
C            print*,'Overflow in MULT'
C            stop
C         end if
         num(len) = carry
      end if

      end
      

      subroutine add(sum,trm,sgntrm,lentrm)
      implicit none
      integer sum(*),trm(*),lentrm,sgntrm
* simple addition of TRM to SUM.  Carry-overs are sorted out
* separately, using occasional calls to REDUCE
      integer i

      do i = 1, lentrm
         sum(i) = sum(i) + sgntrm*trm(i)
      end do

      end

      
      subroutine reduce(sum,maxlen)
      implicit none
      integer maxlen
      integer sum(maxlen)
* every so often it is useful to reduce the size of the terms in SUM
      integer i, carry, temp, base
      parameter (base = 10000)

      carry = 0
      do i = 1, maxlen
         temp = sum(i) + carry
         carry = temp/base
         sum(i) = mod(temp,base)
      end do
      
      end
      

      subroutine final(sum,maxlen,twon)
      implicit none
      integer maxlen, twon
      integer sum(maxlen)
* prepare for output -- make all signs positive and divide by 2^TWON
      integer i, j, temp, carry, base
      parameter (base = 10000)      

* first have to reduce all terms to range -BASE+1 to BASE-1
      call reduce(sum,maxlen)
      
* get them into the range 0 to BASE-1
      carry = 0
      do i = 1, maxlen-1
         if (sum(i) .lt. 0) then
            sum(i) = sum(i) + base
            sum(i+1) = sum(i+1) - 1
         end if
      end do

* do the divisions by 2
      do i = 1, twon
         carry = 0
         do j = maxlen, 1, -1
            temp = sum(j) + base*carry
            sum(j) = temp/2
            carry = mod(temp,2)
         end do
         if (carry.ne.0) print*, 'non-zero remainder in FINAL'
      end do

      end
      

      subroutine prnsum(n,sum,maxlen)
      implicit none
      integer n, maxlen
      integer sum(maxlen)
      
      integer i, lensum
      lensum = maxlen
      do while(sum(lensum) .eq. 0)
         lensum = lensum-1
      end do

      lensum = max(lensum,1)
      print'(a,i2,a,i4,50i1)','L(2,', n, ') = ', sum(lensum),
     &              (mod(sum(i)/1000,10),
     &               mod(sum(i)/100,10),
     &               mod(sum(i)/10,10),
     &               mod(sum(i),10), i = lensum-1, 1, -1)

      end
